/*
** DINA Time daemon.
**
** Watches the user online activity and clears line when time is over.
**
** Based on utmpctl (c) Lev Walkin, 1998, 1999, 2000
**
*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <time.h>
#include <locale.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>
#include <setjmp.h>
#include <err.h>
#include <sysexits.h>

#include <libdina.h>
#include <lgba.h>
#include <strfunc.h>
#include <cost.h>

int slots=0;

int iterate();
int ifprotected();
int clearline(LineEntry *);

time_t	curtime, oldtime;

int CountEm(int);
int Iterator(int, LineEntry *);

typedef struct {
	time_t timestamp;	/* Transition time */
	unsigned int transitions;	/* Transitions */
	short online;	/* 0 - Disconnected, 1 - Connected */
	short protected;	/* 0 - user is subject to time control */

	time_t spent;	/* Money spent by this user */
	time_t i_spent;	/* Money spent by all instances of this user */
	time_t timeleft;	/* What the database says? */
	time_t maxonline;	/* Maximum online time */
	time_t lastsync;	/* How long this entry still not synchronized */

	char **classes;	/* Classes of that user */
	int  classmaster; /* Free the classes upon destoying the structure */

	char username[DI_USERNAME + 1];
} InternalLineEntry;
InternalLineEntry *ILE;

int updateuserinfo(InternalLineEntry *our);
int distributeproperties(InternalLineEntry *, time_t spentQuantum);
time_t getspent(const char *user);

#ifndef CONFIG
#define	CONFIG	"/usr/local/etc/dina.timed.conf"
#endif

char *config=CONFIG;

char *timeleft_attr;
char *maxonline_attr;
slist *prot, *clc;
int daemonmode = 0;

jmp_buf env;
void
onintr(int n) {
	(void)n;
	longjmp(env, 1);
};

void usage();

int
main(int ac, char **av) {
	struct sigaction sa;
	sigset_t sa_mask;
	int les;
	int ch;

	while((ch=getopt(ac, av, "dc:")) != -1)
	switch (ch) {
		case 'c':
			if(config)
				usage();
			config = optarg;
			break;
		case 'd':
			daemonmode = 1;
			break;
		default:
			usage();
	};

	openlog("dina.timed", LOG_CONS, LOG_LOCAL6);

	setlocale(LC_ALL, "");

	if(cfgread(config))
		exit(2);

	if(!(timeleft_attr = cfgget("timeleft")))
		timeleft_attr="timeleft";
	if(!(maxonline_attr = cfgget("maxonlinetime")))
		maxonline_attr = "maxonlinetime";

	prot=sinit(NULL);
	simport(prot, mkarray(cfgget2("protectedclasses"), 0));
	clc=sinit(NULL);
	simport(clc, mkarray(cfgget2("controlclasses"), 0));

	if(lc_init(NULL)) {
		fprintf(stderr, "Error: libcost init error.\n");
		exit(2);
	};

	if((les=lgba_init(NULL)) || (les=lgba_open())) {
		fprintf(stderr, "Error: LGBA error: %s.\n",
			lgba_err2string(les));
		exit(2);
	};

	if(setjmp(env)) {
		syslog(LOG_WARNING, "dina.timed interrupted.");
		return 0;
	};

	sigfillset(&sa_mask);
	sa.sa_handler = onintr;
	sa.sa_mask = sa_mask;
	sa.sa_flags = 0;
	sigaction(SIGHUP, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);
	

	/* Daemonize? */
	if(daemonmode) {
		int fd;

		warnx("All OK. Daemonizing.");

		switch(fork()) {
			case -1:
				err(1, "Daemon mode failed");
				exit(2);
			case 0:
				break;
			default:
				_exit(0);
		};

		setsid();

		if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
			(void)dup2(fd, STDIN_FILENO);
			(void)dup2(fd, STDOUT_FILENO);
			(void)dup2(fd, STDERR_FILENO);
			if (fd > 2)
				(void)close (fd);
		}
	};

	syslog(LOG_INFO, "DINA time control facility started: %d.", getpid());

	iterate();

	return 0;
};

/* Sleep for three seconds before checking the transition state */
#ifndef USLEEP
#define	USLEEP	3000000
#endif

/* Sleep SLEEP*USLEEP before re-counting */
#ifndef SLEEP
#define	SLEEP	20
#endif

/* Refresh account every RESYNC_TIME seconds. */
#ifndef RESYNC_TIME
#define RESYNC_TIME	(5*60)
#endif

int
CountEm(int count) {
	if(count <= 0) return 0;
	if(count <= slots) return (slots=count);

	/* Resize the ILE */
	ILE=(InternalLineEntry *)realloc(ILE,
		sizeof(InternalLineEntry) * count);

	for(; slots < count; slots++)
		memset(ILE+slots, 0, sizeof(InternalLineEntry));

	return slots;
};

int
iterate() {
	size_t saved_transitions = 0;
	int n;

	time(&oldtime);

	while(1) {
		saved_transitions = dina_ctl_get(DCTL_LINE_TRANSITIONS);
		time(&curtime);

		n = iterate_through_lines (NULL, NULL, -1, CountEm, Iterator);
		/*
		** TODO: Perform graceful exit.
		*/
		if(n == -1) {
			syslog(LOG_ERR, "ERROR: Can't iterate through lines!");
		};
	
		oldtime = curtime;

		n=SLEEP;
		while(n--) {
			usleep(USLEEP);
			if(dina_ctl_get(DCTL_LINE_TRANSITIONS) != saved_transitions)
				break;
		};
	};

	return 0;
};

int
Iterator(int slt, LineEntry *le) {
	InternalLineEntry *our = &ILE[slt-1];
	time_t spent;

	/* 1. If in-memory and cached lines are free, get out */
	if(!*le->UserName && !our->online) return 0;

	/* 2. If user has just hanged up */
	if(!*le->UserName) {
		our->online = 0;

		spent = lc_timecost(oldtime,
			(time_t)difftime(curtime, oldtime),
			"online", our->classes);
		our->spent += spent;

		if(our->classmaster) {
			free_values(our->classes);
		};
		our->classes=NULL;
		our->classmaster=0;

		distributeproperties(our, spent);

		syslog(LOG_INFO, "User %s hanged up; time spent online ~ %ld, cost=%ld.", our->username, curtime - our->timestamp, our->spent);
		return 0;
	};

	/* 3. If user connected or changed fastly */
	if (
		le->TimeStamp > our->timestamp
		|| le->Transitions != our->transitions
		|| strncasecmp(our->username, le->UserName, DI_USERNAME)
	) {

		/* Register new user */
		our->online = 0; /* set offline for a while */
		our->spent = 0; /* this user instance has not spent any time */
			/* Get time spent by other instances*/
		our->i_spent = getspent(le->UserName);
		our->timestamp = le->TimeStamp;
		our->transitions = le->Transitions;
		strncpy(our->username, le->UserName, DI_USERNAME);

		updateuserinfo(our);

		/* Add a pair of seconds */
		spent = lc_timecost(our->timestamp,
			(time_t)difftime(curtime, our->timestamp),
			"online", our->classes);

		our->online = 1;

		le->NAS[DI_NAS_ID] = '\0';

		syslog(LOG_INFO, "Connected %s@%s; %ld s. (cost %ld) already; %s.", our->username, le->NAS, curtime - our->timestamp, spent, our->protected?"protected":"clearable");

	} else {	/* User still online */

		spent = lc_timecost(oldtime,
			(time_t)difftime(curtime, oldtime),
			"online", our->classes);

	};

	our->spent += spent;
	our->i_spent += spent;
	distributeproperties(our, spent);

	syslog(LOG_INFO, "User %s is being served, cost %+ld.",
		our->username, spent);

	/* If it's time to refresh a user properties */
	if(!our->protected &&
		( ((our->lastsync - curtime) > RESYNC_TIME)
		|| (our->i_spent >= our->timeleft) )) {

		updateuserinfo(our);

		distributeproperties(our, 0);
		syslog(LOG_DEBUG, "Account %s refilled.", our->username);

		if(our->i_spent >= our->timeleft)
			clearline(le);

	};

	return 0;
};

int
updateuserinfo(InternalLineEntry *our) {
	int n;
	char **vp;

	if((n=load(our->username))) {
		syslog(LOG_WARNING, "Can't load '%s': %s",
			our->username, lgba_err2string(n));
		our->timeleft = 0;
		our->protected = 0;
		our->maxonline = 0;
		our->lastsync = curtime;
		our->classes = (char **)NULL;
		our->classmaster = 0;
		return -1;
	};

	our->timeleft = (time_t)atol(getattr(timeleft_attr));
	our->maxonline = (time_t)atol(getattr("maxonlinetime"));
	if(our->maxonline && (our->maxonline < our->timeleft))
		our->timeleft=our->maxonline;

	vp = getattr2("ObjectClass");
	if(!vp && !copy_values(vp, &our->classes))
		our->classes=NULL;

	our->classmaster = 1;

	our->protected = ifprotected();

	our->lastsync = curtime;

	return 0;
};

/*
** Check if user is protected or must be cleared when resource is over.
*/
int
ifprotected() {
	char **cls;
	char **ours=NULL;

	/*
	** If we cannot get an objectclass attribute,
	** treat user as the non-privoutged
	*/
	if(!(cls=getattr2("objectclass")))
		return 0;
	
	/*
	** Protected classes
	*/
	if(prot)
	for(ours = sarray(prot, 0); *ours; ours++)
	  if(cfind(cls, *ours) >= 0)
		return 1;

	/*
	** Ordinary clearable classes
	*/
	if(clc)
	if(!(ours=sarray(clc, 0)) || !*ours)
		return 1;

	/*
	** User is in the clearable class
	*/
	for(; *ours; ours++) {
	  if(cfind(cls, *ours) != -1)
		return 0;
	};

	syslog(LOG_INFO, "User without clearable class.\n");
	return 1;
};

/*
** Clear user from the specified nas and line using the specified script.
*/
int
clearline(LineEntry *le) {
	static char *script=NULL;
	static char nas[DI_NAS_ID+1];
	static char username[DI_USERNAME+1];

	if(!*le->UserName)
		return -2;

	/*
	** Print default message
	*/
	syslog(LOG_INFO, "User %s: time limit reached. Clearing.",
		le->UserName);


	/*
	** Get the clearing script
	*/
	if(!script) {
		if(!(script=cfgget("clearscript"))) {
			syslog(LOG_WARNING, "Clearing script not defined in config.");
			return -1;
		};
	};

	/* Try to fork */
	switch(fork()) {
	case -1: exit(2);
	case 0: break;
	default: return 0;
	};

	strncpy(nas, le->NAS, DI_NAS_ID);
	nas[DI_NAS_ID] = '\0';
	strncpy(username, le->UserName, DI_USERNAME);
	username[DI_USERNAME] = '\0';

	/*
	** Trying to perform a line clearing.
	*/
	execl(script, "clear", nas, genstr(le->LineNumber), username, NULL);
	syslog(LOG_WARNING, "Cannot execute clearing script.");
	exit(0);
};

/*
** Distribute the recent properties of a particular user (timeleft, protected)
*/
int
distributeproperties(InternalLineEntry *usr, time_t spentQuantum) {
	int i;
	InternalLineEntry *our=ILE;
	int *classmaster=NULL;

	if(!usr) return -1;

	for(i=0; i<slots; i++, our++)
	  if(our->online && usr != our &&
	  !strncasecmp(our->username, usr->username, DI_USERNAME)) {
		our->timeleft = usr->timeleft;
		our->protected = usr->protected;
		our->maxonline = usr->maxonline;
		our->i_spent += spentQuantum;

		if(usr->classmaster) {
			if(our->classmaster) {
				free_values(our->classes);
				our->classes=NULL;
				our->classmaster = 0;
			};
			our->classes = usr->classes;
		} else {
			if(!classmaster)
				classmaster = &our->classmaster;
		};

		our->lastsync = usr->lastsync;

	  };

	if(classmaster)
		*classmaster = 1;

	return 0;
};

/*
** Get the time spent by all instances of a user
*/
time_t
getspent(const char *name) {
	InternalLineEntry *our=ILE;
	int i;
	size_t spent=0;

	if(!name || !*name) return 0;

	for(i=0; i<slots; i++, our++)
	  if(our->online && !strncasecmp(our->username, name, DI_USERNAME))
		spent += our->spent;

	return spent;
};

void
usage() {
	fprintf(stderr, "dina.timed (c) Lev Walkin, 1998, 1999, 2000"
	"\nUsage: dina.timed [-d] [-c <config>]"
	"\n\t-d:\tdetach and become daemon"
	"\n\t-c:\tuse the specified configuration file"
	"\n\t\t(default: " CONFIG ")"
	"\n");
	exit(EX_USAGE);
};
