#ifndef	__LIBDINA_H__
#define __LIBDINA_H__

#include <sys/types.h>

/*
** Line entries in memory represents the physical or logical line on the
** appropriate network access server. The timestamp is always set to the
** last line state change time. The linenumber field represents an
** appropriate server's line. The nas field is set only once upon creation
** of a new line and used to distinguish the line from the distinct access
** servers. Connected and disconnected states are distinguished by checking
** the first byte of the username. If it is set to zero, the line is empty.
*/

#define	DI_NAS_ID	16
#define	DI_USERNAME	32

typedef struct {
	time_t	TimeStamp;	/* Entry change time */
	int	LineNumber;		/* Line number */
	unsigned int SessionID;		/* Optional session identifier */
	unsigned int Transitions;	/* Interface transitions */
	char NAS[DI_NAS_ID];		/* Access server ID */
	char UserName[DI_USERNAME];	/* User name */
} LineEntry;	/* 64 bytes */


typedef struct {
	LineEntry line;	/* LineEntry */
	int	Transition;	/* Global transition counter */
	int empty1;
	int empty2;
	int empty3;
} LineEntryEnvelope;


/*
** Initialize engine in read-only mode. Optional.
** This call should be used before performing any other action.
** Programmer may consider using this function to obtain a non-privileged,
** but serialized access to the shared memory.
*/
int
dina_init_readonly();

/*
** The following functions are used together in the authentication programs,
** performing TACACS+, RADIUS or other authentication mechanizms, including
** a custom ones. Those functions are ONLY needed to authentication programs
** to support DINA, all other functionality will be achieved using the external
** daemons.
*/
/* Declare line as the just-connected */
int
nas_line_set_connected (
	char *nasID,
	int lineNum,
	char *login,
	unsigned int optSessionID	/* You should choose to use it... */
);

/* Declare line down */
int
nas_line_set_disconnected (
	char *nasID,
	int lineNum,
	unsigned int optSessionID	/* lineNum ignored if optSessionID set. */
);



/*
** Iterate through lines.
** Return values:
** -1:	Error case
**  n:	Number of iterated items
*/
int
iterate_through_lines (
	char *optUsername,	/* Select lines with the specified user */
	char *optNasID,		/* Select lines with the specified NAS ID */
	int	optLine,	/* Find line with this line number, -1 otherwise */
		/* Return value: number of items to iterate through */
	int (*optCount)(int Count),
		/* Return values: 0 - OK, any other - STOP */
	int (*optIterateFunc)(int n, LineEntry *line)
);


/*
** This function waits and resurns line transitions, which caused by
** nas_set_line_[dis]connected() invocations. Up to sixteen (15) processes
** may be waiting for a transition event simultaneously.
** Successful event will result in zero exit code and filled LineEntry
** structure.
** Errors are:
**  1: Queue currently empty and NoWait set.
** -1: Common error.
** -2: Can't register the transitions handler (15 processes already listening).
*/
int
dina_next_transition(LineEntryEnvelope *le, int NoWait);



/* Return overall line transitions */
#define	DCTL_LINE_TRANSITIONS	1
#define	DCTL_LINES_COUNT	2
#define	DCTL_IPC_CREATED	3

long
dina_ctl_get(int dctl_Flags);

#endif
